from datetime import datetime

import lal
import numpy as np
from jinja2 import Environment, FileSystemLoader
from pyseobnr.generate_waveform import GenerateWaveform, generate_modes_opt

np.set_printoptions(precision=16)


def get_amp_phase(h):
    amp = np.abs(h)
    phase = np.unwrap(np.angle(h))
    return amp, phase


def sum_sqr_diff(x, y):
    return np.sqrt(np.sum((x - y) ** 2))


def gen_test_data(test_type):
    """
    compute the difference between two waveforms
    and compare to expected value
    """
    m1 = 50.0
    m2 = 30.0
    Mt = m1 + m2
    dt = 1 / 2048.0
    distance = 1.0
    inclination = np.pi / 3.0
    phiRef = 0.0
    approximant = "SEOBNRv5HM"
    s1x = s1y = s2x = s2y = 0.0
    s1z = 0.5
    s2z = 0.1
    f_max = 1024.0
    f_min = 0.0157 / (Mt * np.pi * lal.MTSUN_SI)
    deltaF = 0.125
    params_dict = {
        "mass1": m1,
        "mass2": m2,
        "spin1x": s1x,
        "spin1y": s1y,
        "spin1z": s1z,
        "spin2x": s2x,
        "spin2y": s2y,
        "spin2z": s2z,
        "deltaT": dt,
        "deltaF": deltaF,
        "f22_start": f_min,
        "phi_ref": phiRef,
        "distance": distance,
        "inclination": inclination,
        "f_max": f_max,
        "approximant": approximant,
        "postadiabatic": False,
    }
    wfm_gen = GenerateWaveform(params_dict)  # We call the generator with the parameters
    params_dict2 = params_dict.copy()

    if test_type == "FD":
        # Test differences in Fourier domain, when varying intrinsic params
        hp1, hc1 = wfm_gen.generate_fd_polarizations()
        params_dict2.update({"mass2": 20.0})
        wfm_gen2 = GenerateWaveform(params_dict2)
        hp2, hc2 = wfm_gen2.generate_fd_polarizations()
        # Since we are in FD, hp,hc are *complex*, decompose into amp/phase
        hp1_amp, hp1_phase = get_amp_phase(hp1.data.data)
        hc1_amp, hc1_phase = get_amp_phase(hc1.data.data)
        hp2_amp, hp2_phase = get_amp_phase(hp2.data.data)
        hc2_amp, hc2_phase = get_amp_phase(hc2.data.data)

        hp_amp_diff = sum_sqr_diff(hp1_amp, hp2_amp)
        hp_phase_diff = sum_sqr_diff(hp1_phase, hp2_phase)

        hc_amp_diff = sum_sqr_diff(hc1_amp, hc2_amp)
        hc_phase_diff = sum_sqr_diff(hc1_phase, hc2_phase)

        return hp_amp_diff, hp_phase_diff, hc_amp_diff, hc_phase_diff

    elif test_type == "TD":
        # Test differences in time domain, when varying extrinsic params
        hp1, hc1 = wfm_gen.generate_td_polarizations()
        params_dict2.update({"inclination": 0.17})
        params_dict2.update({"phi_ref": 0.5})
        wfm_gen2 = GenerateWaveform(params_dict2)
        hp2, hc2 = wfm_gen2.generate_td_polarizations()
        # We are in TD, so hp,hc are *real*, compute differences directly
        hp_diff = sum_sqr_diff(hp1.data.data, hp2.data.data)
        hc_diff = sum_sqr_diff(hc1.data.data, hc2.data.data)
        return hp_diff, hc_diff


diffs_TD = np.array2string(np.array(gen_test_data("TD")), separator=",")
diffs_FD = np.array2string(np.array(gen_test_data("FD")), separator=",")


# Load the template
file_loader = FileSystemLoader("./")
env = Environment(loader=file_loader)
template = env.get_template("template_v5HM_tests.jinja")
prog = template.render(
    diffs_TD=diffs_TD,
    diffs_FD=diffs_FD,
    now=datetime.now(),
)
with open("test_SEOBNRv5HM.py", "w") as fw:
    fw.write(prog)



import os
from pathlib import Path

import pandas as pd

import pytest

from pyseobnr.generate_waveform import generate_modes_opt

from .helpers import compare_frames

folder_data = Path(__file__).parent.parent / "data"


@pytest.mark.skipif(
    "CI_TEST_DYNAMIC_REGRESSIONS" not in os.environ,
    reason="regressions on dynamics are for specific systems only",
)
def test_regression_dynamic():
    """
    Checks the dynamic against a full data frame
    """
    q = 5.3
    chi_1 = 0.9
    chi_2 = 0.3
    omega0 = 0.0137  # This is the orbital frequency in geometric units with M=1

    _, _, model = generate_modes_opt(q, chi_1, chi_2, omega0, debug=True)

    frame_hm = pd.DataFrame(
        data=model.dynamics,
        columns="t, r, phi, pr, pphi, H, Omega, Omega_circular".replace(" ", "").split(
            ","
        ),
    )
    frame_hm_reference = pd.read_csv(folder_data / "frame_hm.csv.gz")

    known_differences_percentage = {
        "r": 1,
        "phi": 1,
        "pr": 1,
        "pphi": 1,
        "H": 1,
        "Omega": 1,
        "Omega_circular": 1,
    }

    compare_frames(
        test_frame=frame_hm,
        reference_frame=frame_hm_reference,
        known_differences_percentage=known_differences_percentage,
        time_tolerance_percent=1,
    )



from __future__ import annotations

from typing import Any, Dict, Final, Literal, Tuple, Union, cast, get_args

import lal
import lalsimulation as lalsim
import numpy as np

from .eob.hamiltonian.Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C import (
    Ham_align_a6_apm_AP15_DP23_gaugeL_Tay_C as Ham_aligned_opt,
)
from .eob.hamiltonian.Ham_AvgS2precess_simple_cython_PA_AD import (
    Ham_AvgS2precess_simple_cython_PA_AD as Ham_prec_pa_cy,
)
from .eob.waveform.waveform import SEOBNRv5RRForce
from .models import SEOBNRv5HM
from .models.model import Model

#: Supported approximants
SupportedApproximants = Literal["SEOBNRv5HM", "SEOBNRv5PHM"]


def generate_prec_hpc_opt(
    q: float,
    chi1: np.ndarray,
    chi2: np.ndarray,
    omega_start: float,
    omega_ref: float | None = None,
    settings: Dict[Any, Any] | None = None,
    debug: bool = False,
) -> tuple[np.array, np.array] | tuple[np.array, np.array, Model]:
    """Generate the GW wave polarizations for precessing model in an optimised way. In particular,
    do not compute the inertial frame modes, instead write the polarizations directly summing
    over the co-precessing frame modes

    Args:
        q (float): Mass ratio >=1
        chi1 (np.ndarray): Dimensionless spin of the primary
        chi2 (np.ndarray): Dimensionless spin of the secondary
        omega_start (float): Starting orbital frequency, in geometric units
        omega_ref (float, optional): Reference orbital frequency, in geometric units.
                                     Defaults to None, in which case equals omega_start.
        settings (Dict[Any,Any], optional): Any additional settings to pass to model. Defaults to None.
        debug (bool, optional): Run in debug mode and return the model object. Defaults to False.

    Returns:
        Tuple: Either (time,polarization) or (time,polarizations,model) if debug=True
    """
    if settings is None:
        settings = {"polarizations_from_coprec": True}
    else:
        settings.update(polarizations_from_coprec=True)

    if q < 1.0:
        raise ValueError("mass-ratio has to be positive and with convention q>=1")
    if omega_start < 0:
        raise ValueError("omega_start has to be positive")

    if np.linalg.norm(chi1) > 1 or np.linalg.norm(chi2) > 1:
        raise ValueError("chi1 and chi2 have to respect Kerr limit (|chi|<=1)")

    RR_f = SEOBNRv5RRForce()
    model = SEOBNRv5HM.SEOBNRv5PHM_opt(
        q,
        *chi1,
        *chi2,
        omega_start,
        Ham_prec_pa_cy,
        RR_f,
        omega_ref=omega_ref,
        settings=settings,
    )

    model()
    if debug:
        return model.t, model.hpc, model
    else:
        return model.t, model.hpc


def _check_spins(
    chi1: Union[
        float,
        np.ndarray,
        list,
        tuple[
            float,
            float,
            float,
        ],
    ],
    chi2: Union[
        float,
        np.ndarray,
        list,
        tuple[
            float,
            float,
            float,
        ],
    ],
    approximant: SupportedApproximants,
) -> tuple[float, float, float, float, float, float]:
    chi1 = np.asarray(chi1, dtype=np.float64)
    chi2 = np.asarray(chi2, dtype=np.float64)
    if len(chi1.shape) == 0:
        chi1 = np.array([0, 0, chi1], dtype=np.float64)
    elif chi1.shape[0] > 3:
        raise ValueError("Incorrect number of spin elements.")
    if len(chi2.shape) == 0:
        chi2 = np.array([0, 0, chi2], dtype=np.float64)
    elif chi2.shape[0] > 3:
        raise ValueError("Incorrect number of spin elements.")

    if approximant == "SEOBNRv5HM" and (
        (np.abs(chi1[:2]).max() > 1e-10) or (np.abs(chi2[:2]).max() > 1e-10)
    ):
        raise ValueError(
            "In-plane spin components must be zero for calling non-precessing approximant."
        )
    if np.linalg.norm(chi1) > 1 or np.linalg.norm(chi2) > 1:
        # raise ValueError("chi1 and chi2 have to respect Kerr limit (|chi|<=1)")
        raise ValueError("Dimensionless spin magnitudes cannot be greater than 1!")

    return cast(tuple[float, float, float], tuple(float(_) for _ in chi1)) + cast(
        tuple[float, float, float], tuple(float(_) for _ in chi2)
    )


def generate_modes_opt(
    q: float,
    chi1: Union[float, np.ndarray],
    chi2: Union[float, np.ndarray],
    omega_start: float,
    omega_ref: float = None,
    approximant: SupportedApproximants = "SEOBNRv5HM",
    settings: Dict[Any, Any] = None,
    debug: bool = False,
) -> Tuple:
    """Compute the GW waveform modes for the given configuration and approximant.

    Args:
        q (float): Mass ratio >=1
        chi1 (Union[float,np.ndarray]): Dimensionless spin of primary.
                                        If ``float``, interpreted as z component
        chi2 (Union[float,np.ndarray]): Dimensionless spin of secondary.
                                        If ``float``, interpreted as z component
        omega_start (float): Starting orbital frequency, in geometric units
        omega_ref (float, optional): Reference orbital frequency, in geometric units.
                                     Defaults to None, in which case equals omega_start.
        approximant (SupportedApproximants, optional): The approximant to use. Defaults to "SEOBNRv5HM".
        settings (Dict[Any,Any], optional): Additional settings to pass to model. Defaults to None.
        debug (bool, optional): Run in debug mode . Defaults to False.

    Raises:
        ValueError: If input parameters are not physical
        NotImplementedError: If the approximant requested does not exist

    Returns:
        Tuple: Either (time,dictionary of modes) or (time, dictionary of modes, model) if
        ``debug=True``
    """
    if q < 1.0:
        raise ValueError("mass-ratio has to be positive and with convention q>=1")
    if omega_start < 0:
        raise ValueError("omega_start has to be positive")

    chi1_x, chi1_y, chi1_z, chi2_x, chi2_y, chi2_z = _check_spins(
        chi1, chi2, approximant=approximant
    )

    if approximant == "SEOBNRv5HM":
        RR_f = SEOBNRv5RRForce()
        model = SEOBNRv5HM.SEOBNRv5HM_opt(
            q,
            chi_1=chi1_z,
            chi_2=chi2_z,
            omega0=omega_start,
            H=Ham_aligned_opt,
            RR=RR_f,
            settings=settings,
        )
        model()

    elif approximant == "SEOBNRv5PHM":
        RR_f = SEOBNRv5RRForce()
        model = SEOBNRv5HM.SEOBNRv5PHM_opt(
            q,
            chi1_x=chi1_x,
            chi1_y=chi1_y,
            chi1_z=chi1_z,
            chi2_x=chi2_x,
            chi2_y=chi2_y,
            chi2_z=chi2_z,
            omega_start=omega_start,
            H=Ham_prec_pa_cy,
            RR=RR_f,
            omega_ref=omega_ref,
            settings=settings,
        )
        model()

    else:
        raise NotImplementedError(f"Approximant '{approximant}' is not available")
    if debug:
        return model.t, model.waveform_modes, model
    else:
        return model.t, model.waveform_modes


class GenerateWaveform:
    """
    Class for generating modes, time-domain polarizations and frequency-domain
    polarizations following LAL conventions.
    """

    def __init__(self, parameters):
        """
        Initialize the class with the given ``parameters``.

        ``parameters`` is a dictionary which keys are defined as follow:

        Parameters
        ----------

        float mass1:
            Mass of companion 1, in solar masses - Required
        float mass2:
            Mass of companion 2, in solar masses - Required
        float spin1x:
            x-component of dimensionless spin of companion 1 - Default: 0
        float spin1y:
            y-component of dimensionless spin of companion 1 - Default: 0
        float spin1z:
            z-component of dimensionless spin of companion 1 - Default: 0
        float spin2x:
            x-component of dimensionless spin of companion 2 - Default: 0
        float spin2y:
            y-component of dimensionless spin of companion 2 - Default: 0
        float spin2z:
            z-component of dimensionless spin of companion 2 - Default: 0
        float distance:
            Distance to the source, in Mpc - Default: 100 Mpc
        float inclination:
            Inclination of the source, in radians - Default: 0
        float phi_ref:
            Orbital phase at the reference frequency, in radians - Default: 0
        float f22_start:
            Starting waveform generation frequency, in Hz - Default: 20 Hz
        float f_ref:
            The reference frequency, in Hz - Default: ``f22_start``
        float deltaT:
            Time spacing, in seconds - Default: 1/2048 seconds
        float f_max:
            Maximum frequency, in Hz - Default: 1024 Hz
        float deltaF:
            Frequency spacing, in Hz - Default: 0.125
        list ModeArray:
        list mode_array:
            Mode content (only positive must be specified, e.g ``[(2,2),(2,1)]``).
            Defaults to ``None`` (all modes, see notes below).
        str approximant:

            * ``SEOBNRv5HM`` (default)
            * ``SEOBNRv5PHM``

        Note
        ----

        The default modes are ``(2, 2)``, ``(2, 1)``, ``(3, 3)``, ``(3, 2)``,
        ``(4, 4)`` and ``(4, 3)``. In particular ``(5, 5)`` is not included
        and should be explicitly set through ``ModeArray``.
        """

        self.swap_masses: bool = False
        self.parameters: dict[str, Any] | None = None
        self._validate_parameters(parameters)

    @property
    def model(self):
        if not hasattr(self, "_model"):
            raise ValueError("A model object has not been created!")
        return self._model

    def _validate_parameters(self, parameters):
        if "mass1" not in parameters:
            raise ValueError("mass1 has to be specified!")
        if "mass2" not in parameters:
            raise ValueError("mass2 has to be specified!")

        if parameters["mass1"] > 0 and parameters["mass2"] > 0:
            mass1 = parameters["mass1"]
            mass2 = parameters["mass2"]
        else:
            raise ValueError("Masses have to be positive!")

        Mtot = mass1 + mass2
        if Mtot < 0.001 or Mtot > 1e12:
            raise ValueError("Unreasonable value for total mass, aborting.")

        if mass1 * mass2 / Mtot**2 < 100.0 / (1 + 100) ** 2:
            raise ValueError(
                "Internal function call failed: Input domain error. Model is only valid for systems with "
                "mass-ratio up to 100."
            )

        default_params: Final = {
            "spin1x": 0.0,
            "spin1y": 0.0,
            "spin1z": 0.0,
            "spin2x": 0.0,
            "spin2y": 0.0,
            "spin2z": 0.0,
            "distance": 100.0,
            "inclination": 0.0,
            "phi_ref": 0.0,
            "f22_start": 20.0,
            "deltaT": 1.0 / 2048.0,
            "deltaF": 0.0,
            "ModeArray": None,
            "mode_array": None,
            "approximant": "SEOBNRv5HM",
            "conditioning": 2,
            "polarizations_from_coprec": True,
            "initial_conditions": "adiabatic",
            "initial_conditions_postadiabatic_type": "analytic",
            "postadiabatic": True,
            "postadiabatic_type": "analytic",
            "r_size_input": 12,
        }

        # fills the provided parameters over the default ones
        parameters = default_params | parameters

        if "f_ref" not in parameters.keys():
            parameters["f_ref"] = parameters["f22_start"]

        if parameters["approximant"] not in get_args(SupportedApproximants):
            raise ValueError("Approximant not implemented!")

        # Disable direct polarizations for aligned-spin model
        if parameters["approximant"] == "SEOBNRv5HM":
            parameters["polarizations_from_coprec"] = False

        if "f_max" not in parameters.keys():
            parameters["f_max"] = 0.5 / parameters["deltaT"]

        for param in [
            "spin1x",
            "spin1y",
            "spin1z",
            "spin2x",
            "spin2y",
            "spin2z",
            "distance",
            "inclination",
            "phi_ref",
            "f22_start",
            "f_ref",
            "deltaT",
            "f_max",
            "deltaF",
        ]:
            if not isinstance(parameters[param], float) and not isinstance(
                parameters[param], int
            ):
                raise ValueError(f"{param} has to be a real number!")

        (
            parameters["spin1x"],
            parameters["spin1y"],
            parameters["spin1z"],
            parameters["spin2x"],
            parameters["spin2y"],
            parameters["spin2z"],
        ) = _check_spins(
            chi1=(
                parameters["spin1x"],
                parameters["spin1y"],
                parameters["spin1z"],
            ),
            chi2=(
                parameters["spin2x"],
                parameters["spin2y"],
                parameters["spin2z"],
            ),
            approximant=parameters["approximant"],
        )

        for param in ["f22_start", "f_ref", "f_max", "deltaT", "deltaF", "distance"]:
            if parameters[param] < 0:
                raise ValueError(f"{param} has to be positive!")

        if mass2 > mass1:
            self.swap_masses = True
            aux_spin1 = [
                parameters["spin1x"],
                parameters["spin1y"],
                parameters["spin1z"],
            ]
            parameters["spin1x"], parameters["spin1y"], parameters["spin1z"] = [
                -parameters["spin2x"],
                -parameters["spin2y"],
                parameters["spin2z"],
            ]
            parameters["spin2x"], parameters["spin2y"], parameters["spin2z"] = [
                -aux_spin1[0],
                -aux_spin1[1],
                aux_spin1[2],
            ]
            aux_mass1 = parameters["mass1"]
            parameters["mass1"] = parameters["mass2"]
            parameters["mass2"] = aux_mass1

        if parameters["initial_conditions"] not in ["adiabatic", "postadiabatic"]:
            raise ValueError("Unrecognised setting for initial conditions.")

        if parameters["initial_conditions_postadiabatic_type"] not in [
            "numeric",
            "analytic",
        ]:
            raise ValueError(
                "Unrecognised setting for initial conditions postadiabatic type."
            )

        if parameters["postadiabatic_type"] not in ["numeric", "analytic"]:
            raise ValueError("Unrecognised setting for dynamics postadiabatic type.")

        if parameters["ModeArray"] is not None and parameters["mode_array"] is not None:
            raise ValueError(
                "Only one setting can be specified between ModeArray and mode_array."
            )

        self.parameters = parameters

    def generate_td_modes(self):
        """
        Generate dictionary of positive and negative m modes in physical units.
        """
        fmin, dt = self.parameters["f22_start"], self.parameters["deltaT"]
        f_ref = self.parameters.get("f_ref")
        m1, m2 = self.parameters["mass1"], self.parameters["mass2"]
        Mtot = m1 + m2
        dist = self.parameters["distance"]
        approx: SupportedApproximants = cast(
            SupportedApproximants, self.parameters["approximant"]
        )

        if approx == "SEOBNRv5HM":
            chi1 = self.parameters["spin1z"]
            chi2 = self.parameters["spin2z"]
        elif approx == "SEOBNRv5PHM":
            chi1 = np.array(
                [
                    self.parameters["spin1x"],
                    self.parameters["spin1y"],
                    self.parameters["spin1z"],
                ]
            )
            chi2 = np.array(
                [
                    self.parameters["spin2x"],
                    self.parameters["spin2y"],
                    self.parameters["spin2z"],
                ]
            )

        omega_start = np.pi * fmin * Mtot * lal.MTSUN_SI
        omega_ref = np.pi * f_ref * Mtot * lal.MTSUN_SI
        q = m1 / m2  # Model convention q=m1/m2>=1
        if q < 1.0:
            q = 1 / q

        # Generate internal models, in geometrized units
        settings = {
            "dt": dt,
            "M": Mtot,
            "beta_approx": None,
            "polarizations_from_coprec": False,
        }
        if "postadiabatic" in self.parameters:
            settings.update(postadiabatic=self.parameters["postadiabatic"])

            if "postadiabatic_type" in self.parameters:
                settings.update(
                    postadiabatic_type=self.parameters["postadiabatic_type"]
                )

        if "r_size_input" in self.parameters:
            settings.update(r_size_input=self.parameters["r_size_input"])

        if "initial_conditions" in self.parameters:
            settings.update(initial_conditions=self.parameters["initial_conditions"])

            if "initial_conditions_postadiabatic_type" in self.parameters:
                settings.update(
                    initial_conditions_postadiabatic_type=self.parameters[
                        "initial_conditions_postadiabatic_type"
                    ]
                )

        # Select mode array
        if self.parameters["mode_array"] is not None:
            settings["return_modes"] = self.parameters["mode_array"]
        if self.parameters["ModeArray"] is not None:
            settings["return_modes"] = self.parameters["ModeArray"]

        if "lmax_nyquist" in self.parameters:
            settings.update(lmax_nyquist=self.parameters["lmax_nyquist"])

        settings.update(f_ref=self.parameters["f_ref"])
        times, h, self._model = generate_modes_opt(
            q,
            chi1,
            chi2,
            omega_start,
            approximant=approx,
            omega_ref=omega_ref,
            settings=settings,
            debug=True,
        )

        # Convert to physical units and LAL convention
        Mpc_to_meters = lal.PC_SI * 1e6
        times *= Mtot * lal.MTSUN_SI  # Physical times
        fac = (
            -1 * Mtot * lal.MRSUN_SI / (dist * Mpc_to_meters)
        )  # Minus sign to satisfy LAL convention

        hlm_dict = {}
        for ellm, mode in h.items():
            ell = int(ellm[0])
            if ellm[2] == "-":
                emm = -int(ellm[3])
            else:
                emm = int(ellm[2])
            hlm_dict[(ell, emm)] = fac * mode

        # If aligned-spin model, compute negative modes using equatorial symmetry
        if approx == "SEOBNRv5HM":
            for ellm, mode in h.items():
                ell = int(ellm[0])
                emm = int(ellm[2])
                hlm_dict[(ell, -emm)] = pow(-1, ell) * fac * np.conj(mode)

        # If masses are swapped to satisfy the m1/m2>=1 convention, this implies a
        # pi rotation on the orbital plane, which translates into a minus sign for the odd modes.
        if self.swap_masses is True:
            for ell, emm in hlm_dict.keys():
                if np.abs(emm) % 2 != 0:
                    hlm_dict[(ell, emm)] *= -1.0

        return times, hlm_dict

    def generate_td_polarizations(self):
        """
        Generate time-domain polarizations, returned as LAL REAL8TimeSeries
        """

        incl = self.parameters["inclination"]
        phi = self.parameters["phi_ref"]

        if self.parameters["polarizations_from_coprec"] is False:
            hpc = 0.0
            times, hlm_dict = self.generate_td_modes()
            for ell, emm in hlm_dict:
                hlm = hlm_dict[(ell, emm)]
                ylm = lal.SpinWeightedSphericalHarmonic(
                    incl, np.pi / 2 - phi, -2, ell, emm
                )
                hpc += ylm * hlm

        else:
            fmin, dt = self.parameters["f22_start"], self.parameters["deltaT"]
            f_ref = self.parameters.get("f_ref")
            m1, m2 = self.parameters["mass1"], self.parameters["mass2"]
            Mtot = m1 + m2
            chi1 = np.array(
                [
                    self.parameters["spin1x"],
                    self.parameters["spin1y"],
                    self.parameters["spin1z"],
                ]
            )
            chi2 = np.array(
                [
                    self.parameters["spin2x"],
                    self.parameters["spin2y"],
                    self.parameters["spin2z"],
                ]
            )
            dist = self.parameters["distance"]
            omega_start = np.pi * fmin * Mtot * lal.MTSUN_SI
            omega_ref = np.pi * f_ref * Mtot * lal.MTSUN_SI
            q = m1 / m2
            # Generate internal polarizations, in geometrized units
            settings = {"dt": dt, "M": Mtot, "beta_approx": None}
            if "postadiabatic" in self.parameters:
                settings.update(postadiabatic=self.parameters["postadiabatic"])

                if "postadiabatic_type" in self.parameters:
                    settings.update(
                        postadiabatic_type=self.parameters["postadiabatic_type"]
                    )
            if "initial_conditions" in self.parameters:
                settings.update(
                    initial_conditions=self.parameters["initial_conditions"]
                )

                if "initial_conditions_postadiabatic_type" in self.parameters:
                    settings.update(
                        initial_conditions_postadiabatic_type=self.parameters[
                            "initial_conditions_postadiabatic_type"
                        ]
                    )

            if "r_size_input" in self.parameters:
                settings.update(r_size_input=self.parameters["r_size_input"])

            # Select mode array
            if self.parameters["mode_array"] is not None:
                settings["return_modes"] = self.parameters["mode_array"]
            if self.parameters["ModeArray"] is not None:
                settings["return_modes"] = self.parameters["ModeArray"]

            if "lmax_nyquist" in self.parameters:
                settings.update(lmax_nyquist=self.parameters["lmax_nyquist"])

            settings.update(f_ref=self.parameters["f_ref"])
            Mpc_to_meters = lal.PC_SI * 1e6
            fac = (
                -1 * Mtot * lal.MRSUN_SI / (dist * Mpc_to_meters)
            )  # Minus sign to satisfy LAL convention

            # inclination and phiref for polarizations
            settings.update(inclination=incl)
            # If masses are swapped to satisfy the m1/m2>=1 convention,
            # this implies a pi rotation on the orbital plane.
            if self.swap_masses:
                phi += np.pi
            settings.update(phiref=np.pi / 2 - phi)
            times, hpc, self._model = generate_prec_hpc_opt(
                q,
                chi1,
                chi2,
                omega_start,
                omega_ref=omega_ref,
                settings=settings,
                debug=True,
            )
            hpc *= fac
            times *= Mtot * lal.MTSUN_SI

        hp = np.real(hpc)
        hc = -np.imag(hpc)
        epoch = lal.LIGOTimeGPS(times[0])

        hp_lal = lal.CreateREAL8TimeSeries(
            "hplus", epoch, 0, self.parameters["deltaT"], lal.DimensionlessUnit, len(hp)
        )
        hc_lal = lal.CreateREAL8TimeSeries(
            "hcross",
            epoch,
            0,
            self.parameters["deltaT"],
            lal.DimensionlessUnit,
            len(hp),
        )

        hp_lal.data.data = hp
        hc_lal.data.data = hc

        return hp_lal, hc_lal

    # Procedure as in v4PHM in SimInspiralFD
    def generate_td_polarizations_conditioned_1(self):
        """
        Generate time-domain polarizations, with tappering at the beginning of the waveform,
        returned as LAL REAL8TimeSeries
        """
        hp_lal, hc_lal = self.generate_td_polarizations()
        lalsim.SimInspiralREAL8WaveTaper(hp_lal.data, 1)
        lalsim.SimInspiralREAL8WaveTaper(hc_lal.data, 1)

        return hp_lal, hc_lal

    # General SimInspiralFD procedure, with extra time at the beginning
    def generate_td_polarizations_conditioned_2(self):
        """
        Generate conditioned time-domain polarizations as in SimInspiralTDfromTD routine
        """

        extra_time_fraction = (
            0.1  # fraction of waveform duration to add as extra time for tapering
        )
        extra_cycles = (
            3.0  # more extra time measured in cycles at the starting frequency
        )

        f_min = self.parameters["f22_start"]
        m1 = self.parameters["mass1"]
        m2 = self.parameters["mass2"]
        S1z = self.parameters["spin1z"]
        S2z = self.parameters["spin2z"]
        original_f_min = f_min

        fisco = 1.0 / (pow(9.0, 1.5) * np.pi * (m1 + m2) * lal.MTSUN_SI)
        if f_min > fisco:
            f_min = fisco

        # upper bound on the chirp time starting at f_min
        tchirp = lalsim.SimInspiralChirpTimeBound(
            f_min, m1 * lal.MSUN_SI, m2 * lal.MSUN_SI, S1z, S2z
        )
        # upper bound on the final black hole spin */
        spinkerr = lalsim.SimInspiralFinalBlackHoleSpinBound(S1z, S2z)
        # upper bound on the final plunge, merger, and ringdown time */
        tmerge = lalsim.SimInspiralMergeTimeBound(
            m1 * lal.MSUN_SI, m2 * lal.MSUN_SI
        ) + lalsim.SimInspiralRingdownTimeBound((m1 + m2) * lal.MSUN_SI, spinkerr)

        # extra time to include for all waveforms to take care of situations where the
        # frequency is close to merger (and is sweeping rapidly): this is a few cycles
        # at the low frequency
        textra = extra_cycles / f_min
        # compute a new lower frequency
        fstart = lalsim.SimInspiralChirpStartFrequencyBound(
            (1.0 + extra_time_fraction) * tchirp + tmerge + textra,
            m1 * lal.MSUN_SI,
            m2 * lal.MSUN_SI,
        )

        self.parameters["f22_start"] = fstart
        hp_lal, hc_lal = self.generate_td_polarizations()
        self.parameters["f22_start"] = original_f_min

        # condition the time domain waveform by tapering in the extra time at the
        # beginning and high-pass filtering above original f_min
        lalsim.SimInspiralTDConditionStage1(
            hp_lal, hc_lal, extra_time_fraction * tchirp + textra, original_f_min
        )

        # final tapering at the beginning and at the end to remove filter transients
        # waveform should terminate at a frequency >= Schwarzschild ISCO
        # so taper one cycle at this frequency at the end; should not make
        # any difference to IMR waveforms */
        fisco = 1.0 / (pow(6.0, 1.5) * np.pi * (m1 + m2) * lal.MTSUN_SI)
        lalsim.SimInspiralTDConditionStage2(hp_lal, hc_lal, f_min, fisco)

        return hp_lal, hc_lal

    def generate_fd_polarizations(self):
        """
        Generate Fourier-domain polarizations, returned as LAL COMPLEX16FrequencySeries

        Routine similar to LAL SimInspiralFD.
        """
        # Adjust deltaT depending on sampling rate
        fmax = self.parameters["f_max"]
        f_nyquist = fmax
        deltaF = 0
        if "deltaF" in self.parameters.keys():
            deltaF = self.parameters["deltaF"]

        if deltaF != 0:
            n = int(np.round(fmax / deltaF))
            if n & (n - 1):
                chirplen_exp = np.frexp(n)
                f_nyquist = np.ldexp(1, chirplen_exp[1]) * deltaF

        deltaT = 0.5 / f_nyquist
        self.parameters["deltaT"] = deltaT

        # Generate conditioned TD polarizations
        if self.parameters["conditioning"] == 2:
            hp, hc = self.generate_td_polarizations_conditioned_2()
        else:
            hp, hc = self.generate_td_polarizations_conditioned_1()

        # Adjust signal duration
        if deltaF == 0:
            chirplen = hp.data.length
            chirplen_exp = np.frexp(chirplen)
            chirplen = int(np.ldexp(1, chirplen_exp[1]))
            deltaF = 1.0 / (chirplen * deltaT)
            self.parameters["deltaF"] = deltaF

        else:
            chirplen = int(1.0 / (deltaF * deltaT))

        # resize waveforms to the required length
        lal.ResizeREAL8TimeSeries(hp, hp.data.length - chirplen, chirplen)
        lal.ResizeREAL8TimeSeries(hc, hc.data.length - chirplen, chirplen)

        # FFT - Using LAL routines
        hptilde = lal.CreateCOMPLEX16FrequencySeries(
            "FD H_PLUS",
            hp.epoch,
            0.0,
            deltaF,
            lal.DimensionlessUnit,
            int(chirplen / 2.0 + 1),
        )
        hctilde = lal.CreateCOMPLEX16FrequencySeries(
            "FD H_CROSS",
            hc.epoch,
            0.0,
            deltaF,
            lal.DimensionlessUnit,
            int(chirplen / 2.0 + 1),
        )

        plan = lal.CreateForwardREAL8FFTPlan(chirplen, 0)
        lal.REAL8TimeFreqFFT(hctilde, hc, plan)
        lal.REAL8TimeFreqFFT(hptilde, hp, plan)

        return hptilde, hctilde



import numpy as np
import argparse
import os
import sys
from schwimmbad import JoblibPool
from bilby.core.utils import check_directory_exists_and_if_not_mkdir

sys.path.append(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../calibration"
    )
)
from models import *
from metrics import *
from parameters import parameters_random

from SEOBNRv5_wrapper import SEOBNRv5_wrapper


def perturbation_mismatch(q_A, chi1_A, chi2_A, q_B, chi1_B, chi2_B):
    
    omega0 = 0.015

    model_1 = SEOBNRv5_wrapper(
        q_A,
        chi1_A,
        chi2_A,
        omega0,
    )

    model_2 = SEOBNRv5_wrapper(
        q_B,
        chi1_B,
        chi2_B,
        omega0,
    )
    
    mm = UnfaithfulnessFlat()(model_1, model_2)

    return (
        q_A,
        chi1_A,
        chi2_A,
        mm
    )


def process_one_case(input):
    q_A, chi1_A, chi2_A, q_B, chi1_B, chi2_B = input
    q_A, chi1_A, chi2_A, mm = perturbation_mismatch(q_A, chi1_A, chi2_A, q_B, chi1_B, chi2_B)
    return np.array([q_A, chi1_A, chi2_A, mm])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Perturbation test for a fixed point (q, chi1, chi2)"
    )
    p.add_argument(
        "--points", type=int, help="Number of points", default="5000"
    )
    p.add_argument(
        "--q", type=float, help="Mass-ratio", default="2.0"
    )
    p.add_argument(
        "--chi1", type=float, help="Primary spin", default="0.0"
    )
    p.add_argument(
        "--chi2", type=float, help="Secondary spin", default="0.0"
    )
    p.add_argument(
        "--delta-q", type=float, help="Mass-ratio perturbation", default="1e-9"
    )
    p.add_argument(
        "--delta-chi", type=float, help="Spin perturbation", default="0.01"
    )
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="perturbation_mismatch.dat"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    args = p.parse_args()

    qarr, chi1arr, chi2arr = parameters_random(args.points, args.q - args.delta_q, args.q + args.delta_q, args.chi1 - args.delta_chi, args.chi1 + args.delta_chi, args.chi2 - args.delta_chi, args.chi2 + args.delta_chi)

    lst = [(a, b, c, args.q, args.chi1, args.chi2) for a, b, c in zip(qarr, chi1arr, chi2arr)]

    with JoblibPool(24) as pool:
        all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)

    np.savetxt(args.name, all_means)

    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = './plots'
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        res = np.loadtxt(res_path)

        q = res[:,0]
        chi1 = res[:,1]
        chi2 = res[:,2]
        mm = res[:,3]

        nu = q/(1+q)**2
        m1 = q/(1+q)
        m2 = 1/(1 + q)
        ap = m1*chi1+m2*chi2
        am = m1*chi1-m2*chi2

        # Mismatch across parameter space - (chi1, chi2)
        mm_s, chi1_s, chi2_s = map(list, zip(*sorted(zip(mm, chi1, chi2))))
        plt.scatter(chi1_s, chi2_s, c=mm_s,linewidths=1,norm=matplotlib.colors.LogNorm())
        plt.xlabel('$\chi_1$')
        plt.ylabel('$\chi_2$')
        plt.title(f'$q = {args.q},$'+'$\quad \mathcal{M}_{\mathrm{median}} = $' + f'{np.round(np.median(mm), 6)}')
        cbar=plt.colorbar()
        cbar.set_label('$\mathcal{M}$')
        plt.savefig(f'{plt_dir}/mm_pert_chi1chi2.png', bbox_inches = 'tight', dpi = 300)
        plt.close()

        # Mismatch across parameter space - (ap, am)
        mm_s, ap_s, am_s = map(list, zip(*sorted(zip(mm, ap, am))))
        plt.scatter(ap_s, am_s, c=mm_s,linewidths=1,norm=matplotlib.colors.LogNorm())
        plt.xlabel('$a_{+}$')
        plt.ylabel('$a_{-}$')
        plt.title(f'$q = {args.q},$'+'$\quad \mathcal{M}_{\mathrm{median}} = $' + f'{np.round(np.median(mm), 6)}')
        cbar=plt.colorbar()
        cbar.set_label('$\mathcal{M}$')
        plt.savefig(f'{plt_dir}/mm_pert_apam.png', bbox_inches = 'tight', dpi = 300)
        plt.close()


# cython: language_level=3

cimport cython
from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.waveform.waveform cimport RR_force, RadiationReactionForce
from pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C cimport Hamiltonian_v5PHM_C

cpdef get_rhs_prec(
    double t,
    double[::1] z,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    double m_1,
    double m_2,
    EOBParams params)



from unittest.mock import patch

import numpy as np

from pyseobnr.eob.waveform.compute_hlms import compute_IMR_modes
from pyseobnr.generate_waveform import generate_modes_opt


def test_qnm_rotation_always_positive():
    """Checks we are applying the right rotation to the QNMs

    See https://git.ligo.org/waveforms/software/pyseobnr/-/merge_requests/70 for details
    """

    chi_1 = np.array((-0.0131941810847, 0.0676539715676, -0.646272606477))
    chi_2 = 0
    omega0 = 0.02

    with patch("pyseobnr.models.SEOBNRv5HM.compute_IMR_modes") as p_compute_IMR_modes:
        p_compute_IMR_modes.side_effect = compute_IMR_modes

        generate_modes_opt(
            q=5.33,
            chi1=chi_1,
            chi2=chi_2,
            omega_start=omega0,
            approximant="SEOBNRv5PHM",
        )

        p_compute_IMR_modes.assert_called_once()
        assert p_compute_IMR_modes.call_args.kwargs["qnm_rotation"] > 0

        p_compute_IMR_modes.reset_mock()

        generate_modes_opt(
            q=5.32,
            chi1=chi_1,
            chi2=chi_2,
            omega_start=omega0,
            approximant="SEOBNRv5PHM",
        )

        p_compute_IMR_modes.assert_called_once()
        assert p_compute_IMR_modes.call_args.kwargs["qnm_rotation"] > 0



import datetime
from unittest.mock import patch

import numpy as np
from scipy.interpolate import CubicSpline

import pytest

from pyseobnr.eob.utils.utils import interpolate_dynamics
from pyseobnr.generate_waveform import generate_modes_opt


def interpolate_dynamics_ref(dyn_fine, dt=0.1, peak_omega=None, step_back=250.0):
    """
    Same function as interpolate_dynamics, reference implementation
    """

    res = []

    if peak_omega:
        start = max(peak_omega - step_back, dyn_fine[0, 0])
        t_new = np.arange(start, peak_omega, dt)

    else:
        t_new = np.arange(dyn_fine[0, 0], dyn_fine[-1, 0], dt)

    for i in range(1, dyn_fine.shape[1]):
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, i])
        res.append(intrp(t_new))

    res = np.array(res)
    res = res.T
    return np.c_[t_new, res]


def test_interpolate_dynamics():
    q = 6.855193
    c1 = 0.35492
    c2 = 0.962056
    om = 0.012885

    # we run this once to record the call to interpolate_dynamics
    with patch(
        "pyseobnr.eob.dynamics.integrate_ode.interpolate_dynamics"
    ) as p_interpolate_dynamics_ecc:

        class LocalException(Exception):
            pass

        p_interpolate_dynamics_ecc.side_effect = LocalException

        with pytest.raises(LocalException):
            _, _ = generate_modes_opt(
                q,
                chi1=c1,
                chi2=c2,
                omega_start=om,
                approximant="SEOBNRv5HM",
            )

        p_interpolate_dynamics_ecc.assert_called_once()
        dyn_fine = p_interpolate_dynamics_ecc.call_args_list[0].args[0]
        peak_omega = p_interpolate_dynamics_ecc.call_args_list[0].kwargs["peak_omega"]
        step_back = p_interpolate_dynamics_ecc.call_args_list[0].kwargs["step_back"]

    t1 = datetime.datetime.now()
    for _ in range(1000):
        res1 = interpolate_dynamics_ref(
            dyn_fine=dyn_fine,
            peak_omega=peak_omega,
            step_back=step_back,
        )

    t2 = datetime.datetime.now()
    for _ in range(1000):
        res2 = interpolate_dynamics(
            dyn_fine=dyn_fine,
            peak_omega=peak_omega,
            step_back=step_back,
        )
    t3 = datetime.datetime.now()

    assert (res1 == res2).all()
    assert (t3 - t2).total_seconds() <= 0.7 * (t2 - t1).total_seconds()



import argparse
import importlib
import os
import sys

import gwsurrogate
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool


from pyseobnr.auxiliary.sanity_checks.metrics  import UnfaithfulnessModeByModeLAL
from pyseobnr.auxiliary.external_models  import NRHybSur2dq15Model, NRHybSur3dq8Model,SEOBNRv4HM_LAL
from pyseobnr.generate_waveform import generate_modes_opt

from pyseobnr.auxiliary.sanity_checks.parameters import parameters_random_fast, parameters_random_2D_fast

# Reproducible
seed = 150914

mode_list_v5HM = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3), (5, 5)]
mode_list_v4HM = [(2, 2), (2, 1), (3, 3), (4, 4), (5, 5)]

sur_3dq8 = gwsurrogate.LoadSurrogate("NRHybSur3dq8")
sur_2dq15 = gwsurrogate.LoadSurrogate("NRHybSur2dq15")


def mismatch_NRHybSur(
    q: float,
    chi1: float,
    chi2: float,
    model_name: str,
    approximant_name: str,
    MR_only: bool = False,
):
    # sur = gwsurrogate.LoadSurrogate(model_name)
    if model_name == "NRHybSur2dq15":
        target_model = NRHybSur2dq15Model(
            q,
            chi1,
            chi2,
            0.015,
        )
        sur = sur_2dq15
        mode_list = mode_list_v4HM

    elif model_name == "NRHybSur3dq8":
        target_model = NRHybSur3dq8Model(
            q,
            chi1,
            chi2,
            0.015,
        )
        sur = sur_3dq8
        mode_list = mode_list_v5HM
    target_model(sur)

    if approximant_name == "SEOBNRv5HM":
        _,_,calib_model = generate_modes_opt(q,chi1,chi2,target_model.omega0,debug=True)


    elif approximant_name == "SEOBNRv4HM":
        calib_model = SEOBNRv4HM_LAL(
            target_model.q,
            target_model.chi_1,
            target_model.chi_2,
            target_model.omega0,
        )
        calib_model()
        mode_list = mode_list_v4HM


    masses = np.arange(10, 310, 10)
    unf_settings = {"debug": True, "masses": masses}
    unf = UnfaithfulnessModeByModeLAL(settings=unf_settings)



    mms = []
    for mode in mode_list:
        ell, m = mode
        mm = unf(target_model, calib_model, ell=ell, m=m)
        mms.append(mm)
    return (target_model.q, target_model.chi_1, target_model.chi_2, np.array(mms))


def process_one_case(input):
    q, chi1, chi2, model_name, approximant_name, MR_only = input
    q, chi1, chi2, mm = mismatch_NRHybSur(
        q, chi1, chi2, model_name, approximant_name, MR_only
    )
    # return np.array([q, chi1, chi2, *mm])
    return mm


if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Compute mismatch against NRSur")
    p.add_argument("--points", type=int, help="Number of points", default="5000")
    p.add_argument("--chi-max", type=float, help="Maximum spin", default="0.8")
    p.add_argument("--q-max", type=float, help="Maximum mass-ratio", default="8.0")
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="mismatch"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--model-name",
        type=str,
        help="Name of the surrogate model",
        default="NRHybSur3dq8",
    )
    p.add_argument(
        "--approximant-name",
        type=str,
        help="Name of the approximant",
        default="SEOBNRv5HM",
    )
    p.add_argument(
        "--plots-only",
        action="store_true",
        help="Only generate plots, don't recompute things",
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )
    p.add_argument(
        "--mr-only",
        action="store_true",
        help="Compute unfaithfulness only over the merger-ringdown",
        default=False,
    )

    args = p.parse_args()

    if args.model_name == "NRHybSur2dq15":
        mode_list = mode_list_v4HM
    elif args.model_name == "NRHybSur3dq8":
        mode_list = mode_list_v5HM

    if args.approximant_name == "SEOBNRv4HM":
        mode_list = mode_list_v4HM

    if not args.plots_only:
        # sur = gwsurrogate.LoadSurrogate(args.model_name)

        if args.model_name == "NRHybSur2dq15":
            qarr, chi1arr = parameters_random_2D_fast(
                args.points,
                1.0,
                args.q_max,
                -args.chi_max,
                args.chi_max,
                random_state=seed,
            )
            lst = [
                (q, chi1, 0.0, args.model_name, args.approximant_name, args.mr_only)
                for q, chi1 in zip(qarr, chi1arr)
            ]
            # This model supports non-zero primary spin only
            chi2arr = np.zeros(args.points)

        elif args.model_name == "NRHybSur3dq8":
            qarr, chi1arr, chi2arr = parameters_random_fast(
                args.points,
                1.0,
                args.q_max,
                -args.chi_max,
                args.chi_max,
                -args.chi_max,
                args.chi_max,
                random_state=seed,
            )
            lst = [
                (q, chi1, chi2, args.model_name, args.approximant_name, args.mr_only)
                for q, chi1, chi2 in zip(qarr, chi1arr, chi2arr)
            ]


        pool = Pool(args.n_cpu)
        all_means = pool.map(process_one_case, lst)

        all_means = np.array(all_means)
        # This is an array with shape n cases x p modes x m masses
        mismatches = {}
        for mode in mode_list:
            mismatches[mode] = []
        for i in range(len(all_means)):
            mm_for_case = all_means[i]
            for j, mode in enumerate(mode_list):
                mode_mismatch = mm_for_case[j]
                mismatches[mode].append(mode_mismatch)

        for mode in mode_list:
            ell, m = mode
            np.savetxt(
                f"{args.name}_{args.model_name}_{args.approximant_name}{ell}{m}.dat",
                mismatches[mode],
            )

        np.savetxt(f"parameters_{args.model_name}.dat", np.c_[qarr, chi1arr, chi2arr])
    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = "./plots"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        params = np.genfromtxt(f"parameters_{args.model_name}.dat")
        q = params[:, 0]
        chi1 = params[:, 1]
        chi2 = params[:, 2]

        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1 + m2 * chi2
        am = m1 * chi1 - m2 * chi2
        for mode in mode_list:
            ell, m = mode
            mm_M = np.loadtxt(
                f"{args.name}_{args.model_name}_{args.approximant_name}{ell}{m}.dat"
            )
            if not args.include_all:
                if m % 2 == 1:
                    idx_notEMES = np.where((q > 1.01) | (np.abs(chi1 - chi2) > 0.01))[0]
                    q_pl = q[idx_notEMES]
                    ap_pl = ap[idx_notEMES]
                    mm_M = mm_M[idx_notEMES]
                else:
                    q_pl = q
                    ap_pl = ap
            else:
                q_pl = q
                ap_pl = ap

            if args.mr_only:
                # MR only mismatch uses flat noise curve, does not depend on M
                mm = mm_M
            else:
                M = np.arange(10, 310, 10)

                # maximum of mm_M across total mass for histogram
                mm = []
                for mp in mm_M:
                    mm.append(np.max(mp))
                mm = np.array(mm)

                # Spaghetti plot
                for mp in mm_M:
                    plt.plot(M, mp, color="C0", linewidth=0.5)
                plt.axhline(0.01, ls="--", color="red")
                plt.yscale("log")
                plt.xlabel("M")
                plt.ylabel("$\mathcal{M}$")
                plt.xlim(10, 300)
                plt.savefig(
                    f"{plt_dir}/mm_spaghetti_{args.model_name}_{args.approximant_name}{ell}{m}.png",
                    bbox_inches="tight",
                    dpi=300,
                )
                plt.close()

            # Histogram
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=args.approximant_name + " - " + args.model_name,
            )
            plt.axvline(np.median(mm), c="C0", ls="--")
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.savefig(
                f"{plt_dir}/mm_hist_{args.model_name}_{args.approximant_name}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # CDF
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=100),
                alpha=0.4,
                label=args.approximant_name + " - " + args.model_name,
                cumulative=True,
                density=True,
                histtype="step",
                lw=2,
            )
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.grid(True, which="both", ls=":")
            plt.savefig(
                f"{plt_dir}/mm_cdf_{args.model_name}_{args.approximant_name}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )

            plt.close()

            # Mismatch across parameter space
            mm_s, q_s, ap_s = map(list, zip(*sorted(zip(mm, q_pl, ap_pl))))
            plt.scatter(
                q_s, ap_s, c=mm_s, linewidths=1, norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label("$\mathcal{M}$")
            plt.savefig(
                f"{plt_dir}/mm_scatter_{args.model_name}_{args.approximant_name}{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()



import argparse
import importlib
from mimetypes import MimeTypes
import os
import sys

import lal
import scri
import numpy as np
from bilby.core.utils import check_directory_exists_and_if_not_mkdir
from pathos.multiprocessing import ProcessingPool as Pool

sys.path.append(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../calibration")
)
sys.path.append(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "../../toys/hamiltonian_prototype"
    )
)
sys.path.append(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "../../calibration/calibration_parameter_fits/model_0509")
)
from metrics import *
from models import *

from parameters import parameters_random_fast_prec

from SEOBNRv5PHM_wrapper_0509 import *

# Reproducible
seed = 150914

mode_list = [(2, 2), (2, 1), (3, 3), (3, 2), (4, 4), (4, 3)] # no (5,5) in 7dq4

import gwsurrogate
sur = gwsurrogate.LoadSurrogate('NRSur7dq4')


def package_modes(waveform_modes, ell_min=2, ell_max=5):
    keys = waveform_modes.keys()
    shape = waveform_modes['2,2'].shape
    result = np.zeros((shape[0], 32), dtype=np.complex128)
    i = 0
    for ell in range(ell_min, ell_max + 1):
        for m in range(-ell, ell + 1):

            if str(ell)+','+str(m) not in keys:
                result[:, i] = np.zeros(shape)
            else:
                result[:, i] = waveform_modes[str(ell)+','+str(m)]
            i += 1
    return result


# Set to zero all missing modes in co-precessing frame
def unpack_scri(w):
    result = {}
    for key in w.LM:
        result[f"{key[0]},{key[1]}"] = w.data[
            :, w.index(key[0], key[1])
        ]
    return result



def mismatch_coprecessing(
    q: float,
    chi1x: float,
    chi1y: float,
    chi1z: float,
    chi2x: float,
    chi2y: float,
    chi2z: float,
    approximant_name,
    signal_name,
    NR_file = None,
    ellMax = 5
):

    m1 = q/(1.+q)
    m2 = 1. - m1
    omega0 = 0.022

    Mt = 50.0
    m1 = m1 * Mt
    m2 = m2 * Mt

    f_min = omega0 * (np.pi * (m1 + m2) * lal.MTSUN_SI)

    v5P_wrapper = SEOBNRv5PHMWrapper()

    try:


        if signal_name=="NRSur7dq4":

            model_signal = NRSur7dq4(q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, omega0)
            model_signal(sur)

            imr_full_signal = package_modes(model_signal.waveform_modes)

            # Create a inertial frame scri waveform
            w_signal = scri.WaveformModes(
                dataType=scri.h,
                t=model_signal.t,
                data=imr_full_signal,
                ell_min=2,
                ell_max=5,
                frameType=scri.Inertial,
                r_is_scaled_out=True,
                m_is_scaled_out=True,
            )

            w_signal_cop = w_signal.to_coprecessing_frame()
            modes_signal_cop = unpack_scri(w_signal_cop)

        elif signal_name == 'NR_hdf5':

            model_signal, params_NR = generate_NR_modes(NR_file, ellMax)

            chi1x = params_NR["s1x"]
            chi1y = params_NR["s1y"]
            chi1z = params_NR["s1z"]
            chi2x = params_NR["s2x"]
            chi2y = params_NR["s2y"]
            chi2z = params_NR["s2z"]
            f_min = params_NR["f_min"]
            omega0 = params_NR["omega0"]
            m1 = params_NR['m1']
            m2 = params_NR['m2']
            q = params_NR['q']

        if approximant_name == 'SEOBNRv5PHM':
            params = WaveformParams(
                m1=m1, m2=m2, s1x=chi1x, s1y=chi1y, s1z=chi1z, s2x=chi2x,  s2y=chi2y,  s2z=chi2z,
                f_ref=f_min, f_min = f_min, delta_t=1.0 / 16384, alpha = 0, augm_option = 0#, splines = None
            )

            times, modes_all, model = v5P_wrapper.get_EOB_modes(params)
            coprecessing_modes = model.coprecessing_modes

        elif approximant_name == 'SEOBNRv4PHM':
            settings = {'M':m1+m2 }


            model= SEOBNRv4PHM(q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, omega0,
                              settings = settings)
            model()
            imr_full = package_modes(model.waveform_modes)

            # Create a inertial frame scri waveform
            w = scri.WaveformModes(
                dataType=scri.h,
                t=model.t,
                data=imr_full,
                ell_min=2,
                ell_max=5,
                frameType=scri.Inertial,
                r_is_scaled_out=True,
                m_is_scaled_out=True,
            )

            w_cop = w.to_coprecessing_frame()
            coprecessing_modes = unpack_scri(w_cop)
            model.coprecessing_modes = coprecessing_modes


        elif approximant_name == 'IMRPhenomTPHM':

            settings = {'M':m1+m2 }
            model= IMRPhenomTPHM(q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, omega0,settings= settings)
            model()
            imr_full = package_modes(model.waveform_modes)

            # Create a inertial frame scri waveform
            w = scri.WaveformModes(
                dataType=scri.h,
                t=model.t,
                data=imr_full,
                ell_min=2,
                ell_max=5,
                frameType=scri.Inertial,
                r_is_scaled_out=True,
                m_is_scaled_out=True,
            )

            w_cop = w.to_coprecessing_frame()
            coprecessing_modes = unpack_scri(w_cop)
            model.coprecessing_modes = coprecessing_modes

        for mode in mode_list:
            ell, m = mode
            model.waveform_modes[f'{ell},{m}'] = model.coprecessing_modes[f'{ell},{m}']
            model_signal.waveform_modes[f'{ell},{m}'] = model_signal.coprecessing_modes[f'{ell},{m}']

        masses = np.arange(10, 310, 10)
        unf_settings = {"sigma": 0.001, "debug": True, "masses": masses}
        unf = UnfaithfulnessPhysV4(settings=unf_settings)

        mms = []
        for mode in mode_list:
            ell, m = mode
            mm = unf(model_signal, model, ell=ell, m=m)
            print(f'(ell,m) = {ell},{m}, mm = {np.max(mm)}')
            mms.append(mm)

    except IndexError:
        np.savetxt(f'error_q{np.round(q,5)}.txt', [q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, f_min])
        mms = []
        for mode in mode_list:
            ell, m = mode
            mm = np.ones(30)
            print(f'(ell,m) = {ell},{m}, mm = {np.max(mm)}')
            mms.append(mm)

    return np.array(mms)


def AmpPhysicaltoNRTD(ampphysical, M, dMpc):
    return ampphysical*dMpc*1e6*lal.PC_SI/(lal.C_SI*(M*lal.MTSUN_SI))

def SectotimeM(seconds, M):
    return seconds/(M*lal.MTSUN_SI)

def generate_NR_modes(NR_file:str, ellMax:int):


    try:
        fp = h5py.File(NR_file, "r")
        freq_1M = fp.attrs["f_lower_at_1MSUN"]

        # Add rounding to prevent some errors in the calculation of the symmetric mass ratio (for TEOBResumSe)
        m1_dim = round(fp.attrs["mass1"], 5)
        m2_dim = round(fp.attrs["mass2"], 5)
        e0_signal = round(fp.attrs["eccentricity"], 4)
        mean_anomaly_signal = round(fp.attrs["mean_anomaly"], 4)
        q=m1_dim/m2_dim

        fp.close()
        prefix = os.path.basename(NR_file).split(".")[0]

        LAL_params_signal = lal.CreateDict()
        ma = lalsim.SimInspiralCreateModeArray()

        for ell in range(2, ellMax + 1):
            lalsim.SimInspiralModeArrayActivateAllModesAtL(ma, ell)

        lalsim.SimInspiralWaveformParamsInsertModeArray(LAL_params_signal, ma)
        lalsim.SimInspiralWaveformParamsInsertNumRelData(LAL_params_signal, NR_file)
        (
            s1x,
            s1y,
            s1z,
            s2x,
            s2y,
            s2z,
        ) = lalsim.SimInspiralNRWaveformGetSpinsFromHDF5File(
            0.0, 100, NR_file
        )  # Since we are using f_ref=0, total mass does not matter

        # print(m1_dim, m2_dim, s1x, s1y, s1z, s2x, s2y, s2z)
    except:
        raise print(
            "When using the 'NR_hdf5' approximant, you must provide the NR_file option!"
        )
        exit(-1)

    Mtot = 20.
    distance = 50.0*1e6*lal.PC_SI
    m1 = m1_dim * Mtot
    m2 = m2_dim * Mtot
    fudge_factor = 1.35


    f_min = freq_1M / (m1 + m2)
    omega0 = f_min * (Mtot * lal.MTSUN_SI * np.pi)
    #print(f"f_min = {f_min}, omega0 = {omega0}")
    _, hlm_NR = lalsim.SimInspiralNRWaveformGetHlms(
                1./16384.,
                m1 * lal.MSUN_SI,
                m2 * lal.MSUN_SI,
                distance,
                f_min,
                0.0,
                s1x,
                s1y,
                s1z,
                s2x,
                s2y,
                s2z,
                NR_file,
                ma,
            )


    modes_NR_list =  []

    for ll in range(2,ellMax+1):
        for mm in range(-ll,ll+1):
            modes_NR_list.append([ll,mm])

    modes_NR = {}
    for ll,mm in modes_NR_list:

        key = str(ll)+','+str(mm)
        modes_NR[key] = AmpPhysicaltoNRTD(lalsim.SphHarmTimeSeriesGetMode(hlm_NR, ll,mm).data.data,
                                  m1+ m2, distance/(1e6*lal.PC_SI))

    tmp = lalsim.SphHarmTimeSeriesGetMode(hlm_NR, 2, 2)
    time_NR = tmp.deltaT * np.arange(len(tmp.data.data))
    time_NR = SectotimeM(time_NR, m1 + m2)

    params_NR = {}
    params_NR = {"m1":m1, "m2":m2, 's1x':s1x, 's1y':s1y, 's1z':s1z, 's2x':s2x, 's2y':s2y, 's2z':s2z,
                 "omega0": omega0, 'f_min':f_min, 'q':q }


    alignment_interval_start = 2 # Which peak in Re(h22) is the start
    alignment_interval_end = 12 # Which peak in Re(h22) is the end


    tmin, tmax = get_alignment_interval_from_maxima(
    time_NR,
    np.real(modes_NR["2,2"]),
    tmin=0,
    start=alignment_interval_start,
    end=alignment_interval_end,
    )
    class recast_output:
            def __init__(final, x, fun,tmin,tmax):
                final.t = x
                #final.coprecessing_modes = fun
                final.waveform_modes = fun
                final.tmin = tmin
                final.tmax = tmax


    model_nr = recast_output(time_NR, modes_NR,tmin,tmax)


    modes_nr = model_nr.waveform_modes

    imr_full_nr = package_modes(modes_nr)

    # Create a inertial frame scri waveform
    w_nr = scri.WaveformModes(
        dataType=scri.h,
        t=model_nr.t,
        data=imr_full_nr,
        ell_min=2,
        ell_max=5,
        frameType=scri.Inertial,
        r_is_scaled_out=True,
        m_is_scaled_out=True,
    )

    w_nr_cop = w_nr.to_coprecessing_frame()
    modes_nr_cop = unpack_scri(deepcopy(w_nr_cop))
    model_nr.coprecessing_modes = modes_nr_cop



    return model_nr, params_NR



def process_one_case(input):
    q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, approximant, signal, NR_file, ellMax = input

    mm = mismatch_coprecessing(
        q, chi1x, chi1y, chi1z, chi2x, chi2y, chi2z, approximant, signal,NR_file,ellMax
    )
    return mm


if __name__ == "__main__":
    p = argparse.ArgumentParser(description="Compute mismatch")
    p.add_argument("--points", type=int, help="Number of points", default="1000")
    p.add_argument("--a-max", type=float, help="Maximum spin", default="0.8")
    p.add_argument("--q-max", type=float, help="Maximum mass-ratio", default="4.0")
    p.add_argument(
        "--name", type=str, help="Name of the output file", default="mismatch"
    )
    p.add_argument(
        "--approximant-name", type=str, help="Name of the approximant", default="SEOBNRv5PHM"
    )

    p.add_argument(
        "--signal-name", type=str, help="Name of the signal", default="SEOBNRv5PHM"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument("--n-cpu", type=int, help="Number of cores to use", default=64)
    p.add_argument("--ell-max", type=int, help="Maximum number of l-modes to incude", default=5)
    p.add_argument(
        "--include-all",
        action="store_true",
        help="Include odd m modes close to equal mass",
        default=False,
    )

    p.add_argument("--file_list", type=str, help="List of all the NR files to run on")
    p.add_argument("--nr_dir", type=str, help="Directory with the LVCNR format files", default="/work/sossokine/LVCFormatWaveforms",)


    args = p.parse_args()

    nr_dir = args.nr_dir

    if args.signal_name=='NR_hdf5':

        with open(args.file_list, "r") as fp:
            lst = fp.readlines()
        print(len(lst))
        file_list0 = [lst[i].split('\n')[0] for i in range(len(lst))]

        print(len(lst),file_list0)

        lst = [(1., 0., 0., 0., 0., 0., 0., args.approximant_name,args.signal_name,nr_dir+file_list0[i],args.ell_max) for i in range(len(file_list0))]

        print(f'Generated {len(lst)} NR cases')

    else:
        NR_file=None
        qarr, chi1x_arr, chi1y_arr, chi1z_arr, chi2x_arr, chi2y_arr, chi2z_arr = parameters_random_fast_prec(
            args.points,
            1.0,
            args.q_max,
            0.,
            args.a_max,
            0.,
            args.a_max,
        )

        lst = [(a, b, c, d, e, f, g, args.approximant_name,args.signal_name,NR_file,args.ell_max) for a, b, c, d, e, f, g in zip(qarr, chi1x_arr, chi1y_arr, chi1z_arr, chi2x_arr, chi2y_arr, chi2z_arr)]

        print(f'Generated {args.points} random parameters')

    pool = Pool(args.n_cpu)
    all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)
    # This is an array with shape n cases x p modes x m masses
    mismatches = {}
    for mode in mode_list:
        mismatches[mode] = []
    for i in range(len(all_means)):
        mm_for_case = all_means[i]
        for j, mode in enumerate(mode_list):
            mode_mismatch = mm_for_case[j]
            mismatches[mode].append(mode_mismatch)

    for mode in mode_list:
        ell, m = mode
        np.savetxt(
            f"{args.name}_{args.approximant_name}_{args.signal_name}_copr_{ell}{m}.dat",
            mismatches[mode],
        )

    np.savetxt(f"parameters_{args.approximant_name}_{args.signal_name}_copr.dat", np.c_[qarr, chi1x_arr, chi1y_arr, chi1z_arr, chi2x_arr, chi2y_arr, chi2z_arr])
    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = "./plots"
        check_directory_exists_and_if_not_mkdir(plt_dir)

        res_path = args.name
        params = np.genfromtxt(f"parameters_{args.approximant_name}_{args.signal_name}_copr.dat")
        q = params[:, 0]
        chi1_x = params[:, 1]
        chi1_y = params[:, 2]
        chi1_z = params[:, 3]
        chi2_x = params[:, 4]
        chi2_y = params[:, 5]
        chi2_z = params[:, 6]


        nu = q / (1 + q) ** 2
        m1 = q / (1 + q)
        m2 = 1 / (1 + q)
        ap = m1 * chi1_z + m2 * chi2_z
        am = m1 * chi1_z - m2 * chi2_z

        for mode in mode_list:
            ell, m = mode
            mm_M = np.loadtxt(
                f"{args.name}_{args.approximant_name}_{args.signal_name}_copr_{ell}{m}.dat"
            )
            # maximum of mm_M across total mass for histogram
            mm = []
            for mp in mm_M:
                mm.append(np.max(mp))
            mm = np.array(mm)

            # Do not plot cases with errors
            idx = np.where(mm<1.0)[0]
            q_pl = q[idx]
            ap_pl = ap[idx]
            mm_M = mm_M[idx]
            mm = mm[idx]

            M = np.arange(10, 310, 10)


            # Spaghetti plot
            for mp in mm_M:
                plt.plot(M, mp, color="C0", linewidth=0.5)
            plt.axhline(0.01, ls="--", color="red")
            plt.yscale("log")
            plt.xlabel("M")
            plt.ylabel("$\mathcal{M}$")
            plt.xlim(10, 300)
            plt.savefig(
                f"{plt_dir}/mm_spaghetti_{args.approximant_name}_{args.signal_name}_copr_{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # Histogram
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=50),
                alpha=0.4,
                label=f'{args.approximant_name}-{args.signal_name} (coprecessing frame)',
            )
            plt.axvline(np.median(mm), c="C0", ls="--")
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.savefig(
                f"{plt_dir}/mm_hist_{args.approximant_name}-{args.signal_name}_copr_{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()

            # CDF
            plt.hist(
                mm,
                bins=np.logspace(start=np.log10(0.00001), stop=np.log10(1.0), num=100),
                alpha=0.4,
                label=f'{args.approximant_name}-{args.signal_name}',
                cumulative=True,
                density=True,
                histtype="step",
                lw=2,
            )
            plt.legend(loc="best")
            plt.gca().set_xscale("log")
            plt.xlabel("$\mathcal{M}_{\mathrm{Max}}$")
            plt.title(
                "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),6)}"
            )
            plt.grid(True, which="both", ls=":")
            plt.savefig(
                f"{plt_dir}/mm_cdf_{args.approximant_name}-{args.signal_name}_copr_{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )

            plt.close()

            # Mismatch across parameter space
            mm_s, q_s, ap_s = map(list, zip(*sorted(zip(mm, q_pl, ap_pl))))
            plt.scatter(
                q_s, ap_s, c=mm_s, linewidths=1, norm=matplotlib.colors.LogNorm()
            )
            plt.ylabel("$\chi_{\mathrm{eff}}$")
            plt.xlabel("$q$")
            cbar = plt.colorbar()
            cbar.set_label("$\mathcal{M}$")
            plt.savefig(
                f"{plt_dir}/mm_scatter{args.approximant_name}-{args.signal_name}_copr_{ell}{m}.png",
                bbox_inches="tight",
                dpi=300,
            )
            plt.close()



#!/usr/bin/env python
"""
Contains functions associated with evolving the equations of motion
"""

import numpy as np
import pygsl_lite.errno as errno
import pygsl_lite.odeiv2 as odeiv2
from numba import jit
from scipy.interpolate import CubicSpline

from ..utils.utils import interpolate_dynamics, iterative_refinement
from .initial_conditions_aligned_opt import computeIC_opt
from .rhs_aligned import augment_dynamics, get_rhs

step = odeiv2.pygsl_lite_odeiv2_step
_control = odeiv2.pygsl_lite_odeiv2_control
evolve = odeiv2.pygsl_lite_odeiv2_evolve


class control_y_new(_control):
    def __init__(self, eps_abs, eps_rel):
        a_y = 1
        a_dydt = 1
        super().__init__(eps_abs, eps_rel, a_y, a_dydt, None)


@jit(nopython=True)
def h_max(r):
    return 1


def ODE_system_RHS_opt(t: float, z: np.ndarray, args) -> np.ndarray:
    """Return the dynamics equations for aligned-spin systems

    Args:
        t (float): The current time
        z (np.array): The dynamics variables, stored as (q,p)
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use. Must have same signature as Hamiltonian
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        np.array: The dynamics equations, including RR
    """
    return get_rhs(t, z, *args)


def compute_dynamics_opt(
    omega0,
    H,
    RR,
    chi_1,
    chi_2,
    m_1,
    m_2,
    rtol=1e-11,
    atol=1e-12,
    backend="solve_ivp",
    params=None,
    step_back=100,
    max_step=0.1,
    min_step=1.0e-9,
    y_init=None,
    r_stop=None,
):
    """
    Main function to integrate the dynamics
    The RHS of the equations are given in Eq(2) of [Buades2021]_ .
    See rhs_aligned.pyx for more details.
    Uses GSL Dormand-Prince 8th order integrator.
    Args:
        omega0 (float): initial orbital frequency
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use. Must have same signature as Hamiltonian
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary
        rtol (float): relative tolerance
        atol (float): absolute tolerance
        step_back (float): step back for the start of the fine dynamics
        r_stop (float): minimum final separation for the dynamics

    Returns:
        np.array, np.array: coarse and fine dynamics arrays
    """

    sys = odeiv2.system(
        ODE_system_RHS_opt, None, 4, [H, RR, chi_1, chi_2, m_1, m_2, params]
    )

    T = odeiv2.step_rk8pd
    s = step(T, 4)
    c = control_y_new(atol, rtol)
    e = evolve(4)

    t = 0
    t1 = 2.0e9
    if r_stop < 0:
        r_stop = 1.4

    if y_init is None:
        r0, pphi0, pr0 = computeIC_opt(
            omega0, H, RR, chi_1, chi_2, m_1, m_2, params=params
        )
        y0 = np.array([r0, 0.0, pr0, pphi0])
    else:
        y0 = y_init.copy()
    y = y0
    if y_init is None:
        h = 2 * np.pi / omega0 / 5
    else:
        h = 0.5
    omega_previous = omega0
    res_gsl = []
    ts = []

    ts.append(0.0)
    res_gsl.append(y)

    p_circ = np.zeros(2)
    peak_omega = False
    peak_pr = False
    while t < t1:
        # Take a step
        status, t, h, y = e.apply(c, s, sys, t, t1, h, y)
        if status != errno.GSL_SUCCESS:
            print("break status", status)
            break
        # Compute the error for the step controller
        e.get_yerr()

        # Append the last step
        res_gsl.append(y)
        ts.append(t)

        # Compute the RHS after the step is done

        r = y[0]

        # Check if the proposed step is larger than the maximum timestep
        # h_mx = h_max(r)
        # h = h / h_mx

        # Handle termination conditions
        if r <= 6:
            deriv = ODE_system_RHS_opt(t, y, [H, RR, chi_1, chi_2, m_1, m_2, params])
            drdt = deriv[0]
            omega = deriv[1]
            dprdt = deriv[2]
            """
            h_small = np.max((0.01,2*np.pi/(2.*omega) / (1 + np.exp(-(r - 4) / 0.13))))
            if h > h_small:
                h = h_small
            """

            if omega < omega_previous:
                peak_omega = True
                break
            if drdt > 0:
                break
            if dprdt > 0:
                peak_pr = True
                break
            if r <= r_stop:
                break
            if r < 3:
                q_vec = y[:2]
                p_circ[1] = y[-1]
                omega_circ = H.omega(q_vec, p_circ, chi_1, chi_2, m_1, m_2)
                if omega_circ > 1:
                    break
            omega_previous = omega

    ts = np.array(ts)
    dyn = np.array(res_gsl)

    if peak_omega:
        t_desired = ts[-1] - step_back - 50
    else:
        t_desired = ts[-1] - step_back

    idx_close = np.argmin(np.abs(ts - t_desired))
    if ts[idx_close] > t_desired:
        idx_close -= 1

    # Guard against the case where when using PA dynamics,
    # there is less than step_back time between the start
    # of the ODE integration and the end of the dynamics
    # In that case make the fine dynamics be _all_ dynamics
    # except the 1st element
    if t_desired < ts[1]:
        idx_close = 1
        step_back = ts[-1] - ts[idx_close]
    dyn_coarse = np.c_[ts[:idx_close], dyn[:idx_close]]
    dyn_fine = np.c_[ts[idx_close:], dyn[idx_close:]]

    # print(f"t_desired={t_desired}")
    # print(f"len(dyn_coarse)={len(dyn_coarse)}")
    # print(f"len(dyn_fine)={len(dyn_fine)}")
    # print(f"End time: {ts[-1]}")
    # print(f"idx_close:{idx_close}, t[idx_close]={ts[idx_close]}")

    dyn_coarse = augment_dynamics(dyn_coarse, chi_1, chi_2, m_1, m_2, H)
    dyn_fine = augment_dynamics(dyn_fine, chi_1, chi_2, m_1, m_2, H)
    t_peak = None
    if peak_omega:
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, -2])
        left = dyn_fine[0, 0]
        right = dyn_fine[-1, 0]
        t_peak = iterative_refinement(intrp.derivative(), [left, right])

    if peak_pr:
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, 3])
        left = dyn_fine[-1, 0] - 10
        right = dyn_fine[-1, 0]
        t_peak = iterative_refinement(intrp.derivative(), [left, right], pr=True)

    dyn_fine = interpolate_dynamics(
        dyn_fine[:, :-3], peak_omega=t_peak, step_back=step_back
    )
    dyn_fine = augment_dynamics(dyn_fine, chi_1, chi_2, m_1, m_2, H)

    return dyn_coarse, dyn_fine



import argparse
import importlib
import os,sys
from mimetypes import MimeTypes
import lal
import numpy as np


import matplotlib.pyplot as plt
from pathlib import Path
from pathos.multiprocessing import ProcessingPool as Pool
from pyseobnr.generate_waveform import generate_modes_opt

from pycbc.psd.analytical import aLIGOZeroDetHighPower, aLIGOZeroDetHighPowerGWINC
from pycbc.types import TimeSeries
from pycbc.filter import make_frequency_series
from pycbc.filter.matchedfilter import optimized_match
from pycbc.waveform.utils import taper_timeseries

from typing import Dict, Union, Tuple
def combine_modes(
    iota: float, phi: float, modes_dict: Dict
) -> Tuple[np.array, np.array]:
    """Combine modes to compute the waveform polarizations in the direction
    (iota,np.pi/2-phi)

    Args:
        iota (float): Inclination angle (rad)
        phi (float): Azimuthal angle(rad)
        modes_dict (Dict): Dictionary containing the modes, either time of frequency-domain

    Returns:
        np.array: Waveform in the given direction
    """
    sm = 0.0
    for key in modes_dict.keys():
        #print(key)
        ell, m = [int(x) for x in key.split(",")]
        Ylm0 = lal.SpinWeightedSphericalHarmonic(iota, np.pi / 2 - phi, -2, ell, m)
        sm += Ylm0 * modes_dict[key]

    return np.real(sm), -np.imag(sm)


def timeMtoSec(timeM, mt:float):
    return timeM*mt*lal.MTSUN_SI


def ampNRtoPhysicalTD(ampNR, mt:float,distance:float):
    return ampNR*(lal.C_SI*mt*lal.MTSUN_SI)/distance


def generate_v5PHM_waveform(m1:float, m2:float,
                            s1x:float, s1y:float, s1z:float,
                            s2x:float, s2y:float, s2z:float,
                            omega_ref:float,
                            omega_start:float,
                            iota:float, phi:float, distance:float,
                            delta_t:float, approx:str,
                            ell_max: int = 4,
                            initial_conditions = "adiabatic",
                            initial_conditions_postadiabatic_type = "analytic",
                            ):

    mtotal = m1+m2
    q = m1/m2
    #omega0 = f_min * (np.pi * (m1 + m2) * lal.MTSUN_SI)

    chi_1 = [s1x, s1y, s1z]
    chi_2 = [s2x, s2y, s2z]

    settings = {
            'ell_max':ell_max,'beta_approx':None,'M':mtotal,"dt":delta_t,
            "initial_conditions" : initial_conditions,
            "initial_conditions_postadiabatic_type" : initial_conditions_postadiabatic_type,
            }

    if approx == 'SEOBNRv5PHM_PA':
        settings["postadiabatic"] = True
        settings["postadiabatic_type"] = "analytic"

        #print(f"params = {q,chi_1,chi_2,omega0}")
        #print(f"settings = {settings}")
        time, modes, model = generate_modes_opt(q,chi_1,chi_2,omega_start,approximant='SEOBNRv5PHM',
                           omega_ref= omega_ref, debug=True,settings=settings)
        modes_dict = {}
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                modes_dict["{},{}".format(ell, m)] = modes["{},{}".format(ell, m)]

    elif approx == 'SEOBNRv5PHM_PA_num':
        settings["postadiabatic"] = True
        settings["postadiabatic_type"] = "numeric"

        #print(f"params = {q,chi_1,chi_2,omega0}")
        #print(f"settings = {settings}")
        time, modes, model = generate_modes_opt(q,chi_1,chi_2,omega_start,approximant='SEOBNRv5PHM',
                                         omega_ref= omega_ref, debug=True,settings=settings)
        modes_dict = {}
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                modes_dict["{},{}".format(ell, m)] = modes["{},{}".format(ell, m)]


    elif approx == 'SEOBNRv5PHM':
        settings["postadiabatic"] = False

        time, modes, model = generate_modes_opt(q,chi_1,chi_2,omega_start,approximant=approx,
                                         omega_ref= omega_ref,
                                         debug=True,settings=settings)
        modes_dict = {}
        for ell in range(2, ell_max + 1):
            for m in range(-ell, ell + 1):
                modes_dict["{},{}".format(ell, m)] = modes["{},{}".format(ell, m)]

    elif approx == 'SEOBNRv5HM':

        time, modes = generate_modes_opt(q,chi_1[-1],chi_2[-1],omega_start,approximant=approx,
                                debug=False,settings=settings)

        modes_dict = {}
        for key in modes.keys():
            lm_tag = key.split(',')
            ell,m = lm_tag
            ll = float(ell)
            modes_dict["{},{}".format(ell, m)] = modes[key] # Minus sign to have same convention as v5PHM
            modes_dict["{},-{}".format(ell,m)] = pow(-1.,ll) * np.conjugate(modes[key]) # Minus sign to have same convention as v5PHM


    else:
        raise NotImplementedError



    hp_NR, hc_NR = combine_modes(iota, phi, modes_dict)

    t_s = timeMtoSec(time, mtotal)
    hp = ampNRtoPhysicalTD(hp_NR,mtotal,distance)
    hc = ampNRtoPhysicalTD(hc_NR,mtotal,distance)


    # Taper
    hp_td = TimeSeries(hp, delta_t=delta_t)
    hc_td = TimeSeries(hc, delta_t=delta_t)
    hp_td = taper_timeseries(hp_td, tapermethod="startend")
    hc_td = taper_timeseries(hc_td, tapermethod="startend")

    return time, modes_dict, hp_td, hc_td, model



def pa_mismatch_prec(m1,m2, s1x,s1y,s1z,s2x,s2y,s2z, iota_s,ell_max=4,initial_conditions="adiabatic",omega0 = 0.02):


    chi1 = [s1x,s1y,s1z]
    chi2 = [s2x,s2y,s2z]

    Mt = m1+m2
    q=m1/m2

    # Do not test sub-solar mass binaries
    if m2 < 1:

        m2 = 1.
        m1 = q*m2
        Mt = m1+m2


    if q<1:
        q=1

    phi_s = 0.

    omega_ref = omega0
    omega_start = omega_ref

    distance = 1e6*lal.PC_SI
    delta_t = 1./16384.0 
    f_min = omega_start / (Mt * lal.MTSUN_SI * np.pi)


    # Generate TD polatizations of SEOBNRv5PHM

    approx = 'SEOBNRv5PHM'
    # non-PA evolution
    time, modes, hp_td, hc_td,_ = generate_v5PHM_waveform(m1, m2,
                                                        s1x, s1y, s1z,
                                                        s2x, s2y, s2z,
                                                        omega_ref,
                                                        omega_start,
                                                        iota_s, phi_s, distance,
                                                        delta_t, approx, ell_max = ell_max,
                                                        initial_conditions = initial_conditions,
                                                        initial_conditions_postadiabatic_type = "analytic",
                                                        )


    approx = 'SEOBNRv5PHM_PA'
    # PA evolution
    time_pa, modes_pa, hp_td_pa, hc_td_pa,_ = generate_v5PHM_waveform(m1, m2,
                                                        s1x, s1y, s1z,
                                                        s2x, s2y, s2z,
                                                        omega_ref,
                                                        omega_start,
                                                        iota_s, phi_s, distance,
                                                        delta_t, approx, ell_max = ell_max,
                                                        initial_conditions = "postadiabatic",
                                                        initial_conditions_postadiabatic_type = "analytic",
                                                        )

    # Pad zeros
    N = max(len(hp_td_pa), len(hc_td))
    pad = int(2 ** (np.floor(np.log2(N)) + 2))
    hp_td_pa.resize(pad)
    hc_td_pa.resize(pad)
    hp_td.resize(pad)
    hc_td.resize(pad)

    # Perform the Fourier Transform
    hp_pa = make_frequency_series(hp_td_pa)
    hc_pa = make_frequency_series(hc_td_pa)

    # Perform the Fourier Transform
    hp = make_frequency_series(hp_td)
    hc = make_frequency_series(hc_td)


    amp_strain = abs(hp.data-1j*hc.data)
    idx_max = np.argmax(amp_strain)
    fpeak_nonpa = hp.get_sample_frequencies()[idx_max]

    amp_pa_strain = abs(hp_pa.data-1j*hc_pa.data)
    idx_max = np.argmax(amp_pa_strain)
    fpeak_pa = hp_pa.get_sample_frequencies()[idx_max]

    fpeak = max([fpeak_nonpa,fpeak_pa])
    f_min = 2.0*fpeak_pa

    # Generate PSD
    if f_min<10:
        f_low_phys = 10.
    else:
        f_low_phys = f_min

    f_high_phys = 2048.

    psd = aLIGOZeroDetHighPowerGWINC(len(hp), hp.delta_f, f_low_phys)

    # Compute match for hplus
    mm_hp = optimized_match(hp,
          hp_pa,
          psd,
          low_frequency_cutoff=f_low_phys,
          high_frequency_cutoff=f_high_phys
          )[0]


    # Compute match for hcross
    mm_hc = optimized_match(hc,
        hc_pa,
        psd,
        low_frequency_cutoff=f_low_phys,
        high_frequency_cutoff=f_high_phys
        )[0]

    # Take the mean
    #print(f"mm_hp = {mm_hp}, mm_hc = {mm_hc}")
    mm_mean  = 1.-np.mean([mm_hp,mm_hc])
    #print(f"mm = {mm_mean}")
    """except:

        print(
            f"Error for the following parameters: q = {q}, chi1 = {chi1}, chi2 = {chi2}, Mt = {Mt}, iota_s = {iota_s}"
        )
        mm_mean = -1
        pass
    """



    if mm_mean>0.001:


        plt.figure(figsize = (10,10/1.618))

        mode_list = ['2,2','2,1','3,3','3,2','4,4','4,3']#,'5,5']

        amp22 = np.abs(modes['2,2'])
        idx_max = np.argmax(amp22)
        tmax22 = time[idx_max]

        amp22_pa = np.abs(modes_pa['2,2'])
        idx_max = np.argmax(amp22_pa)
        tmax22_pa = time_pa[idx_max]

        for mode in mode_list:
            plt.plot(time-tmax22, np.real(modes[mode]),label=str(mode))
            plt.plot(time_pa-tmax22_pa, np.real(modes_pa[mode]),ls = '--')


        plt.xlabel("Time (M)")
        plt.ylabel(r"$\|h_{\ell m}|$")
        plt.xlim(-300,100)
        plt.legend(loc=3)
        plt.title(f'MM = {np.round(mm_mean,7)} - q{np.round(q,3)}_s{np.round(chi1,3)}_s{np.round(chi2,3)} - Mt = {np.round(Mt,3)}')
        Path('./plots').mkdir(parents=True, exist_ok=True)
        plt.savefig(f'./plots/pa_q{np.round(q,3)}_s{np.round(chi1,3)}_s{np.round(chi2,3)}_M{np.round(Mt,3)}.png', dpi = 200)
        plt.clf()

    res = [m1,m2, *chi1,*chi2, mm_mean,iota_s]

    return (m1,m2, *chi1,*chi2,iota_s, mm_mean)



def process_one_case(input):
    m1,m2,  s1x,s1y,s1z,s2x,s2y,s2z, iota_s, ell_max,initial_conditions,omega0 = input

    m1,m2,  s1x,s1y,s1z,s2x,s2y,s2z, iota_s, mm = pa_mismatch_prec(m1,m2, s1x,s1y,s1z,s2x,s2y,s2z, iota_s, ell_max =ell_max,initial_conditions=initial_conditions,omega0 = omega0)
    return np.array([m1,m2,s1x,s1y,s1z,s2x,s2y,s2z, iota_s, mm])


if __name__ == "__main__":

    p = argparse.ArgumentParser(
        description="Test post-adiabatic for the full polarizations"
    )
    p.add_argument(
        "--points", type=int, help="Number of points", default="2000"
    )
    p.add_argument(
        "--plots", action="store_true", help="Make diagnostic plots", default=True
    )
    p.add_argument(
        "--ncores", type=int, help="Number of cores to use", default="64"
    )
    p.add_argument(
        "--q-max", type=float, help="Maximum mass-ratio", default="100.0"
    )
    p.add_argument(
        "--M-max", type=float, help="Maximum total mass", default="10.0"
    )
    p.add_argument(
        "--M-min", type=float, help="Maximum total mass", default="300.0"
    )
    p.add_argument(
        "--ell-max", type=int, help="Maximum l-mode", default="4"
    )
    p.add_argument(
        "--initial-conditions", type=str, help="Type of initial conditions to use for the non-PA evolution", default="adiabatic"
    )
    p.add_argument(
        "--omega0", type=float, help="Starting orbital frequeny in geometric units", default="0.02"
    )

    p.add_argument(
        "--aligned", type=str, help="If set then run aligned-spin configuration", default="False"
    )
    args = p.parse_args()

    seed=150914
    N = args.points
    a_max = 0.99
    ell_max = args.ell_max
    omega0 = args.omega0
    initial_conditions = args.initial_conditions
    as_case = args.aligned

    np.random.seed(seed)
    q = np.random.uniform(1., args.q_max, N)
    #m1 = np.random.uniform(args.M_min,args.M_max,N)
    np.random.seed(seed-1)
    mtotal = np.random.uniform(args.M_min, args.M_max, N)

    m1 = q/(1.+q)*mtotal
    m2 = 1./(1.+q)*mtotal

    if as_case == "True":

        np.random.seed(seed+999998)
        chi1z = np.random.uniform(-a_max, a_max, N)
        np.random.seed(seed+999997)
        chi2z = np.random.uniform(-a_max, a_max, N)
        chi1x = np.zeros(N)
        chi1y = np.zeros(N)
        chi2x = np.zeros(N)
        chi2y = np.zeros(N)

    else:
        np.random.seed(seed+999998)
        a1 = np.random.uniform(0., a_max, N)
        np.random.seed(seed+999997)
        a2 = np.random.uniform(0., a_max, N)
        np.random.seed(seed+999996)
        theta1 = np.random.uniform(0,np.pi,N)
        np.random.seed(seed+999995)
        theta2 = np.random.uniform(0,np.pi,N)
        np.random.seed(seed+999994)
        phi1 = np.random.uniform(0,2*np.pi,N)
        np.random.seed(seed+999993)
        phi2 = np.random.uniform(0,2*np.pi,N)

        chi1x = a1*np.sin(theta1)*np.cos(phi1)
        chi1y = a1*np.sin(theta1)*np.sin(phi1)
        chi1z = a1*np.cos(theta1)

        chi2x = a2*np.sin(theta2)*np.cos(phi2)
        chi2y = a2*np.sin(theta2)*np.sin(phi2)
        chi2z = a2*np.cos(theta2)

    np.random.seed(seed+999992)
    cos_iota = np.random.uniform(-1.0, 1.0, args.points)
    iota_arr = np.arccos(cos_iota)

    ell_max_arr =  np.full(len(iota_arr), ell_max, dtype=int)
    ic_arr =  np.full(len(iota_arr), initial_conditions)
    omega0_arr = np.full(len(iota_arr), omega0)
    lst = [(a, b, c, d,e,f,g,h,i,j,k,l) for a, b, c, d,e,f,g,h,i,j,k,l in zip(m1,m2,chi1x,chi1y,chi1z,chi2x,chi2y,chi2z,iota_arr,ell_max_arr,ic_arr,omega0_arr)]

    with Pool(args.ncores) as pool:
        all_means = pool.map(process_one_case, lst)

    all_means = np.array(all_means)

    np.savetxt(f"pa_mismatch.dat", all_means)

    ### Plots ###

    if args.plots:

        import matplotlib
        import matplotlib.pyplot as plt

        plt_dir = './plots'
        Path(plt_dir).mkdir(parents=True, exist_ok=True)

        res_path = f"pa_mismatch.dat"
        res = np.loadtxt(res_path)

        m1 = res[:,0]
        m2 = res[:,1]
        q = m1/m2

        chi1x = res[:,2]
        chi1y = res[:,3]
        chi1z = res[:,4]

        chi2x = res[:,5]
        chi2y = res[:,6]
        chi2z = res[:,7]

        iota = res[:,8]
        mm = res[:,9]

        nu = q/(1+q)**2
        m1 = q/(1+q)
        m2 = 1/(1 + q)
        m_total = m1 + m2
        ap = (m1*chi1z+m2*chi2z)/m_total
        am = (m1*chi1z-m2*chi2z)/m_total

        # Mismatch across parameter space
        mm_s, q_s, ap_s, am_s = map(list, zip(*sorted(zip(mm, q, ap, am))))
        plt.scatter(q_s, ap_s, c=mm_s,linewidths=1,norm=matplotlib.colors.LogNorm())
        plt.xlabel('$q$')
        plt.ylabel('$a_{+}$')
        plt.title('$\mathcal{M}_{\mathrm{median}} = $' + f'{np.round(np.median(mm), 7)}')
        cbar=plt.colorbar()
        cbar.set_label('$\mathcal{M}$')
        plt.savefig(f'{plt_dir}/mm_pa_apam.png', bbox_inches = 'tight', dpi = 300)
        plt.close()


        # Histogram

        fig, ax = plt.subplots(figsize=(14,10),dpi=250)
        ax.hist(
            mm,
            bins=np.logspace(start=np.log10(1e-15), stop=np.log10(1.0), num=100),
            alpha=0.4,
            label='SEOBNRv5PHM non-PA - SEOBNRv5PHM PA',
            histtype="step",
            fill=False
        )

        ax.axvline(np.median(mm), c="C0", ls="--")
        ax.legend(loc="best")
        ax.set_xlabel("$\mathcal{M}_{\mathrm{Max}}$",fontsize=30)
        ax.set_ylabel("Count",fontsize=30)
        ax.set_title(
            "$\mathcal{M}_{\mathrm{median}} = $" + f"{np.round(np.median(mm),10)}",fontsize=25
        )


        ax.set_xscale('log')

        size=30
        ax.tick_params(axis='x', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='x', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='major', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        ax.tick_params(axis='y', which='minor', pad=10,width=2,length=5,size=7, labelsize=size,direction='in')
        plt.tight_layout()
        plt.savefig(f"{plt_dir}/mm_hist_v5PHM_pa_test.png",bbox_inches="tight",dpi=300,)
        plt.close()



from .EOB_fits import (  # noqa
    EOBCalculateNQCCoefficients_freeattach,
    EOBCalculateRDAmplitudeConstraintedCoefficient1,
    EOBCalculateRDAmplitudeConstraintedCoefficient2,
    EOBCalculateRDPhaseConstraintedCoefficient1,
    EOBNonQCCorrection,
    compute_QNM,
)
from .fits_Hamiltonian import NR_deltaT, NR_deltaT_NS, a6_NS, dSO  # noqa
from .GSF_fits import GSF_amplitude_fits  # noqa
from .IV_fits import InputValueFits  # noqa
from .MR_fits import MergerRingdownFits  # noqa



import numpy as np
from scipy.interpolate import CubicSpline
from scipy.signal import argrelmin


def iterative_refinement(f, interval, levels=2, dt_initial=0.1, pr=False):
    """
    Attempts to find the peak of Omega/pr iteratively.
    Needed to ensure accurate attachment when the attachment point is the last point of the dynamics.

    Args:
        f (PPoly): derivative of the splined dynamics
        interval (list): interval where to look for the peak
        levels (int): number of iterations (should be greater than 2)
        dt_initial (float): initial time step
        pr (bool): whether the end of the dynamics is due to a peak of pr instead of Omega

    Returns:
        np.array: interpolated dynamics array

    """

    assert levels > 1

    left = interval[0]
    right = interval[1]
    result = None

    for n in range(1, levels + 1):
        dt = dt_initial / (10**n)
        t_fine = np.arange(interval[0], interval[1], dt)
        deriv = np.abs(f(t_fine))

        mins = argrelmin(deriv, order=3)[0]
        if len(mins) > 0:
            result = t_fine[mins[0]]

            interval = max(result - 10 * dt, left), min(result + 10 * dt, right)

        else:
            if pr:
                return interval[-1]
            else:
                return (interval[0] + interval[-1]) / 2

    assert result is not None
    return result


def interpolate_dynamics(dyn_fine, dt=0.1, peak_omega=None, step_back=250.0):
    """
    Interpolate the dynamics to a finer grid.
    This replaces stepping back that was used in older EOB models.

    Args:
        dyn_fine (np.array): dynamics array
        dt (float): time step to which to interpolate
        peak_omega (float): position of the peak (stopping condition for the dynamics)
        step_back (float): step back relative to the end of the dynamics

    Returns:
        np.array: interpolated dynamics array

    """

    if peak_omega:
        start = max(peak_omega - step_back, dyn_fine[0, 0])
        t_new = np.arange(start, peak_omega, dt)

    else:
        t_new = np.arange(dyn_fine[0, 0], dyn_fine[-1, 0], dt)

    intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, 1:])
    res = intrp(t_new)

    return np.c_[t_new, res]



#!/usr/bin/env python
"""This file contains the default settings for various models
Each default setting is a *function* the resturns a dictionary.
To be used by models in models.py
"""

from typing import Any, Dict


def default_unfaitfulness_mode_by_mode_settings() -> Dict[Any, Any]:
    settings = dict(
        sigma=0.001,  # the sigma to use in likelihood
    )
    return settings
